/*
 *
 *  * Copyright (C) 2014 Numatica Corporation.
 *  * All Rights Reserved.
 *  *
 *  * NOTICE:  All information contained herein is, and remains
 *  * the property of Numatica Corporation and its suppliers,
 *  * if any.  The intellectual and technical concepts contained
 *  * herein are proprietary to Numatica Corporation
 *  * and its suppliers and may be covered by U.S. and Foreign Patents,
 *  * patents in process, and are protected by trade secret or copyright law.
 *  * Dissemination of this information or reproduction of this material
 *  * is strictly forbidden unless prior written permission is obtained
 *  * from Numatica Corporation.
 *
 */

package com.numatica.insightmagic;

import com.hazelcast.config.*;
import com.hazelcast.core.Hazelcast;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.MapLoader;
import com.hazelcast.core.MapStore;
import com.numatica.insightmagic.geolocation.store.GeoStore;
import com.numatica.insightmagic.interaction.store.UserActionStore;
import com.numatica.insightmagic.session.store.SessionStore;
import com.numatica.insightmagic.user.store.UserAccountStore;
import com.numatica.insightmagic.user.store.UserPreferencesStore;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;

import javax.annotation.Resource;
import java.util.Arrays;
import java.util.List;


@Configuration
public class HazelcastConfiguration {

    public static final String USER_ACCOUNT_MAP = "userAccountMap";
    public static final String USER_PREFS_MAP = "userPreferencesMap";
    public static final String SESSION_MAP = "sessionMap";

    public static final String GEO_MAP = "geoMap";
    public static final String USER_ACTION_HISTORY_MAP = "userActionHistoryMap";

    public static final String USER_ACTION_QUEUE = "userActionQueue";

    // number of seconds that refresh tokens live
    private static final int REFRESH_TTL = 3600;
    public static final int NO_OF_BACKUPS = 1;

    @Autowired
    private DataGridSettings hzSettings;

    @Autowired
    @Resource
    @Qualifier("userAccountStore")
    private UserAccountStore userAccountStore;

    @Autowired
    @Resource
    @Qualifier("userPreferencesStore")
    private UserPreferencesStore userPreferencesStore;

    @Autowired
    @Resource
    @Qualifier("sessionStore")
    private SessionStore sessionStore;

    @Autowired
    @Resource
    @Qualifier("geoStore")
    private GeoStore geoStore;

    @Autowired
    @Resource
    @Qualifier("userActionStore")
    private UserActionStore userActionStore;


    @Bean(name = "hazelcastInstance")
    HazelcastInstance hazelcast() {

        final Config config = new Config();

        config.addMapConfig(initializeMap(USER_ACCOUNT_MAP, new String[]{"emailAddress"}, userAccountStore, MapConfig.EvictionPolicy.LRU, 0));
        config.addMapConfig(initializeMap(USER_PREFS_MAP, null, userPreferencesStore, MapConfig.EvictionPolicy.LRU, 0));
        config.addMapConfig(initializeNearCacheMap(SESSION_MAP, Arrays.asList("userId"), sessionStore, MapConfig.EvictionPolicy.NONE, 0));
        config.addMapConfig(initializeMap(USER_ACTION_HISTORY_MAP, new String[]{"userSessionId", "actionCode", "ipAddress", "timestamp"},
                userActionStore, MapConfig.EvictionPolicy.LRU, 0));

        config.addMapConfig(initializeMap(GEO_MAP, new String[]{"ipStartNum", "ipEndNum", "city", "countryCode", "stateOrProvince"},
                geoStore, MapConfig.EvictionPolicy.NONE, 0));

        config.addQueueConfig(initializeQueue(USER_ACTION_QUEUE));

        if (hzSettings.isManagementCenterEnabled()) {
            config.getManagementCenterConfig().setUrl(hzSettings.getManagementCenterURL()).setEnabled(true);
        }

        NetworkConfig network = config.getNetworkConfig();
        JoinConfig join = network.getJoin();
        if (hzSettings.isMulticast()) {
            join.getMulticastConfig().setEnabled(true);
        } else {
            List<String> tcpIpMembers = hzSettings.getTcpIpMembers();
            if (!tcpIpMembers.isEmpty()) {
                join.getMulticastConfig().setEnabled(false);
                join.getTcpIpConfig().setEnabled(true);
                tcpIpMembers.stream()
                    .filter(member -> null != member)
                    .forEach(member -> join.getTcpIpConfig().addMember(member));
            }
        }
        List<String> interfaces = hzSettings.getInterfaces();
        if (!interfaces.isEmpty()) {
            network.getInterfaces().setEnabled(true);
            interfaces.stream()
                  .filter(n -> null != n)
                  .forEach(n -> network.getInterfaces().addInterface(n));
        }

        initializeQueue(USER_ACTION_QUEUE);

    return Hazelcast.newHazelcastInstance(config);
}

    private QueueConfig initializeQueue(String queueName) {
        QueueConfig queueConfig = new QueueConfig();
        queueConfig.setStatisticsEnabled( true );
        queueConfig.setName( queueName );
        return queueConfig;
    }


    private MapConfig initializeMap(final String mapName, String[] indices,
                                    final MapLoader mapStore,
                                    final MapConfig.EvictionPolicy policy,
                                    final int ttl) {
        final MapConfig mapConfig = new MapConfig();
        mapConfig.setName(mapName);
        if (indices != null) {
            for (String index : indices) {
                mapConfig.addMapIndexConfig(new MapIndexConfig(index, true));
            }
        }

        mapConfig.setBackupCount(NO_OF_BACKUPS);
        mapConfig.setEvictionPolicy(policy != null ? policy : MapConfig.EvictionPolicy.NONE);
        mapConfig.setEvictionPercentage(25);
        mapConfig.setTimeToLiveSeconds(ttl);
        //mapCfg.setMergePolicy("")

        if (null != mapStore) {
            final MapStoreConfig mapStoreConfig = new MapStoreConfig();
            mapStoreConfig.setEnabled(true);
            //mapStoreCfg.setWriteDelaySeconds(10);
            mapStoreConfig.setImplementation(mapStore);
            mapStoreConfig.setInitialLoadMode(MapStoreConfig.InitialLoadMode.LAZY);
            mapConfig.setMapStoreConfig(mapStoreConfig);
        }

        return mapConfig;
    }


    private MapConfig initializeNearCacheMap(final String mapName, List<String> indices,
                                             final MapStore mapStore,
                                             final MapConfig.EvictionPolicy policy,
                                             final int ttl) {
        final MapConfig mapConfig = new MapConfig();
        mapConfig.setName(mapName);
        //sessionCfg.setBackupCount( 1 );
        mapConfig.setEvictionPolicy(policy != null ? policy : MapConfig.EvictionPolicy.NONE);
        mapConfig.setEvictionPercentage(25);
        mapConfig.setTimeToLiveSeconds(ttl);
        //mapCfg.setMergePolicy("")

        if (indices != null) {
            for (String index : indices) {
                mapConfig.addMapIndexConfig(new MapIndexConfig(index, true));
            }
        }

        NearCacheConfig nearCacheConfig = new NearCacheConfig();
        nearCacheConfig.setEvictionPolicy("NONE");
        nearCacheConfig.setInMemoryFormat(InMemoryFormat.OBJECT);
        nearCacheConfig.setCacheLocalEntries(true); // this enables local caching
        mapConfig.setNearCacheConfig(nearCacheConfig);

        if (null != mapStore) {
            final MapStoreConfig mapStoreConfig = new MapStoreConfig();
            mapStoreConfig.setEnabled(true);
            //mapStoreCfg.setWriteDelaySeconds(10);
            mapStoreConfig.setImplementation(mapStore);
            mapStoreConfig.setInitialLoadMode(MapStoreConfig.InitialLoadMode.LAZY);
            mapConfig.setMapStoreConfig(mapStoreConfig);
        }

        return mapConfig;
    }

}
